/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002 Julian Missig
 */

#include "GabberApp.hh"

#include <config.h>

#include "Environment.hh"
#include "ConfigPaths.hh"
#include "PrefsInterface.hh" // for LoginDlg
#include "ModuleManager.hh"
#include "S10nRequestDlg.hh"
#include "GabberWin.hh"
#include "ResourceManager.hh"
#include "LogManager.hh"
#include "PlatformFuncs.hh"
#include "S5B.hh"
#include "FileTransferDlg.hh"
#include "StreamInitiation.hh"
#include "FileTransferManager.hh"
#include "FileTransferRecvDlg.hh"

#include "AutoAway.hh"
#include "StatusIcon.hh"

#include "intl.h"

#include <jabberoo/presence.hh>

#include <gtkmm/box.h>
#include <gtkmm/dialog.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/fileselection.h>
#include <gtkmm/image.h>

using namespace Gabber;
using namespace jabberoo;

// -------------------------------------------------------------------
//
// Ye Olde Gabber App - The Hub
//
// -------------------------------------------------------------------
GabberApp::GabberApp(int argc, char** argv)
     : _GtkMain(argc, argv), _winMain(NULL)
{
     cout << ENV_VARS.package << " " << ENV_VARS.version << endl;

     _log_manager = new LogManager();
     _configurator = create_configurator();

     // Instiatiate objects
     _resource_manager = new ResourceManager();
     _module_manager = new ModuleManager();
     _ft_manager = new FileTransferManager();

     // Filter for agents
    _session->discoDB().signal_new_item.connect(
        SigC::slot(*this, &GabberApp::on_discodb_filter));

     // Session signals
     _session->evtTransmitXML.connect(slot(*this, &GabberApp::on_XML_transmit));
     _session->evtRecvXML.connect(slot(*this, &GabberApp::on_XML_recv));
//     _session->evtIQ.connect(slot(*this, &GabberApp::on_session_iq));
     _session->evtAuthError.connect(slot(*this, &GabberApp::on_session_auth_error));
     _session->evtPresenceRequest.connect(slot(*this, &GabberApp::on_presence_request));
     _session->evtPresenceUnsubscribed.connect(slot(*this, &GabberApp::on_presence_unsubscribed));
//     _session->evtUnknownPacket.connect(slot(*this, &GabberApp::on_session_unknown_packet));
//     _session->evtPresence.connect(slot(*this, &GabberApp::on_session_presence));
//     _session->evtMyPresence.connect(slot(*this, &GabberApp::on_session_my_presence));
    _si_query = _session->registerXPath(
        "/iq[@type='set']/si[@xmlns='http://jabber.org/protocol/si']", 
        SigC::slot(*this, &GabberApp::on_stream_initiation));

     // Add all the features we have
     _session->addFeature("http://jabber.org/protocol/si");
     _session->addFeature("http://jabber.org/protocol/si/profile/file-transfer");
     _session->addFeature("http://jabber.org/protocol/bytestreams");

     // Check if we should do first-time wizard
     // if (config ...
     // otherwise, do they want to autoconnect?
     if (_configurator->get_bool(Keys::acct.autoconnect))
     {
          if (_configurator->get_bool(Keys::acct.savepassword))
          {
               init_win();
               set_password(_configurator->get_string(Keys::acct.password));
               connect();
          }
          else
          {
               get_password();
          }
     }
     else
     {
          // ok, just give them the login dialog
          LoginDlg::execute();
     }
}

GabberApp::~GabberApp()
{
     _session->unregisterXPath(_si_query);
     delete _ft_manager;
     delete _resource_manager;
     delete _module_manager;
     delete _log_manager;
}

void GabberApp::run()
{
     // Let the gnome app start processing messages
     Gtk::Main::run();
}

void GabberApp::get_password()
{
     Gtk::Dialog dlg;
     dlg.set_title(_("Password required"));
     dlg.set_has_separator(false);
     dlg.set_resizable(false);
     Gtk::Button btnLogIn("_Log In", true);
     btnLogIn.property_can_default().set_value(true);
     Gtk::Button btnCancel(Gtk::StockID("gtk-cancel"));
     dlg.add_action_widget(btnCancel, Gtk::RESPONSE_CANCEL);
     dlg.add_action_widget(btnLogIn, Gtk::RESPONSE_OK);
     dlg.set_default_response(Gtk::RESPONSE_OK);
     btnCancel.show();
     btnLogIn.show();
     
     Gtk::HBox hb(false, 12);
     hb.set_border_width(6);
     dlg.get_vbox()->pack_start(hb);
     hb.show();
     
     Gtk::Image di(Gtk::StockID("gtk-dialog-question"), Gtk::ICON_SIZE_DIALOG);
     hb.pack_start(di, Gtk::PACK_SHRINK);
     di.show();
     
     Gtk::VBox vb(false, 6);
     hb.pack_start(vb);
     vb.show();
     
     Gtk::Label lblMain;
     lblMain.set_line_wrap(true);
     lblMain.set_use_markup(true);
     lblMain.set_markup("<span weight='bold' size='larger'>"
                        + string(_("Password required")) + "</span>\n\n"
                        + Util::substitute(_("Enter your password for JabberID <b>%s@%s</b>."),
                                           _configurator->get_string(Keys::acct.username),
                                           _configurator->get_string(Keys::acct.server)));
     vb.pack_start(lblMain);
     lblMain.show();
     
     Gtk::HBox passbox(false, 3);
     Gtk::Label lblPass(_("_Password:"), true);
     Gtk::Entry entPass;
     lblPass.set_selectable(false);
     lblPass.set_mnemonic_widget(entPass);
     entPass.set_visibility(false);
     entPass.set_activates_default(true);
     passbox.pack_start(lblPass);
     passbox.pack_start(entPass);
     lblPass.show();
     entPass.show();
     vb.pack_start(passbox);
     passbox.show();
     dlg.show();
     
     int resp = dlg.run();
     dlg.hide();
     if (resp == Gtk::RESPONSE_OK)
     {
          init_win();
          set_password(entPass.get_text());
          connect();
     }
     else
     {
          LoginDlg::execute();
     }
}

void GabberApp::init_win()
{
    // Setup all our connection info
    _connection_info.server = _configurator->get_string(Keys::acct.server);
    _connection_info.port = _configurator->get_int(Keys::acct.port);
    _connection_info.username = _configurator->get_string(Keys::acct.username);
    _connection_info.resource = _configurator->get_string(Keys::acct.resource);
    // This is now handled by the set_password function so it doesn't have to be
    // saved in GConf
    //_connection_info.password = _configurator->get_string(Keys::acct.password);
    _connection_info.ssl = _configurator->get_bool(Keys::acct.ssl);
    _connection_info.reconnect = _configurator->get_bool(Keys::acct.reconnect);

    // If we don't have an existing main window
    if (_winMain == NULL)
    {
         // Create one
         _winMain  = new GabberWin(*this);
         _autoaway = new AutoAway(_winMain); // hook up autoaway to this window
         _status_icon = new StatusIcon(_winMain);
         printf("Loading modules!\n");
         _module_manager->loadAll();
    }
}

void GabberApp::clean_disconnect()
{
    // Send out status
    *_session << jabberoo::Presence("", jabberoo::Presence::ptUnavailable,
    jabberoo::Presence::stOffline, _("Logged Out"));

    // Try disconnecting Session.. and if that fails, the Transmitter
    if (!_session->disconnect())
        _transmitter->disconnect();
}

void GabberApp::quit()
{
     clean_disconnect();

     // Cleanup GUI stuff before the msg loop goes boom
     delete _autoaway;
     delete _status_icon;
     delete _winMain;

     // Stop gtk/gnome message loop
     Gtk::Main::quit();
}

// -------------------------------------
// Session event handlers
// -------------------------------------
void GabberApp::on_XML_transmit(const char* XML)
{
#ifdef SESSION_DEBUG
    _log_manager->debug() << jutil::getTimeStamp() << " send " 
        << XML << endl;
#endif // SESSION_DEBUG
}

void GabberApp::on_XML_recv(const char* XML)
{
#ifdef SESSION_DEBUG
    _log_manager->debug() << jutil::getTimeStamp() << " recv " 
        << XML << endl;
#endif // SESSION_DEBUG
}

void GabberApp::on_session_auth_error(int ErrorCode, const char* ErrorMsg)
{
     // TODO: handle these errors differently when in registration phase
     // perhaps disconnect this handler and use a separate one?

     Gtk::MessageDialog* dlg;
     
     switch (ErrorCode)
     {
          case 401:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in: Authorization failed</b>\nYour username or password may be incorrect."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("Authorization failed"));
               dlg->run();
               dlg->hide();
               get_password();
	       delete dlg;
               return;
          case 402:
          case 407:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in: External registration required</b>\nThis server requires some sort of external registration or payment. Try looking at the server's web site."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("External registration required"));
               dlg->run();
               dlg->hide();
               LoginDlg::execute();
	       delete dlg;
               return;
          case 406:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in: Required information not provided</b>\nAdditional authentication is required. Make sure the password is not blank. It is possible Gabber does not support the authentication required for this server."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("Required information not provided"));
               dlg->run();
               dlg->hide();
               LoginDlg::execute();
	       delete dlg;
               return;
          case 409:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in: Resource conflict</b>\nYou have another connection using the same resource. Choose a different resource."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("Resource conflict"));
               dlg->run();
               dlg->hide();
               LoginDlg::execute();
	       delete dlg;
               return;
          case 500:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in: Internal server error</b>\nThis server encountered an error while attempting to authenticate. Try again later. It is possible Gabber does not support the authentication required for this server."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("Internal server error"));
               dlg->run();
               dlg->hide();
               LoginDlg::execute();
	       delete dlg;
               return;
          default:
               dlg = new Gtk::MessageDialog(_("<b>Could not log in</b>\nAn unknown error occurred while attempting to log in."), Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg->set_title(_("Internal server error"));
               dlg->run();
               dlg->hide();
               LoginDlg::execute();
	       delete dlg;
               return;
     }
}

void GabberApp::on_presence_request(const jabberoo::Presence& pres)
{
     std::string jid = pres.getFrom();
     std::string reason = pres.getStatus();
     jabberoo::Presence::Type type = pres.getType();

     if (type == jabberoo::Presence::ptSubRequest)
     {
          // If this has already been queued
          // or if we want messages auto-displayed and are currently free for chat or available
          // we can view it now, else queue
          if (pres.getBaseElement().cmpAttrib("gabber:queued", "true")
              || (_configurator->get_bool(Keys::interface.messages_autodisplay)
                  && (_currPresence.getShow() == jabberoo::Presence::stChat
                      || _currPresence.getShow() == jabberoo::Presence::stOnline)
                  )
              )
          {
               // Other user wants access to our presence
               S10nRequestDlg::display(*this, *_winMain->getGtkWindow(), jid, reason);
          }
          else
          {
               _pqueue->push(new judo::Element(pres.getBaseElement()), "presence-s10n.png", "PresenceRequest");
          }
     }
     else if (type == jabberoo::Presence::ptUnsubRequest)
     {
          // Other user no longer wants our presence
          jabberoo::Presence p(jid, jabberoo::Presence::ptUnsubscribed);
          *_session << p;
     }
}

void GabberApp::on_presence_unsubscribed(const jabberoo::Presence& pres)
{
     // Other user denies our presence viewing abilities
     // TODO: How do we tell if we /requested/ unsubscription
     // or if they just did this to us?
     // ANSWER: if they're still in our roster, then they just did this to us.
     // b& for sass
     cerr << "Received a presence unsubscribed packet from " << pres.getFrom() << endl;
     cerr << "TODO: How do we tell if we /requested/ unsubscription or if they just did this to us? See GabberApp.cc" << endl;
}

void GabberApp::on_stream_initiation(const judo::Element& elem)
{
    // XXX See if this is a type of profile we'll support, if not we can still
    // do a generic save dialog.
    const judo::Element* si_elem = elem.findElement("si");
    SI* si = new SI(*si_elem);

    if ( si->getProfile() && si->getProfile()->getNamespace() == 
            "http://jabber.org/protocol/si/profile/file-transfer" )
    {
        FeatureNegotiation* fneg = si->getFeatureNegotiation();
        if (fneg && !fneg->check("stream-method", 
                "http://jabber.org/protocol/bytestreams"))
        {
            // We don't have a valid stream, send an error
            judo::Element* err = SI::buildInvalidStreamsNode();
            err->putAttrib("to", elem.getAttrib("from"));
            err->putAttrib("id", elem.getAttrib("id"));
            err->putAttrib("type", "error");
            *_session << err->toString().c_str();
            delete err;
            delete si;
            return;
        }

        // Figure out the name to save it as
        // XXX should be a confirmation dialog
        printf("SI Node: %s\n", elem.toString().c_str());
        FileTransferRecvDlg* dlg = Gtk::manage(
            new FileTransferRecvDlg(elem.getAttrib("from"), 
                                    elem.getAttrib("id"), si));
        dlg->show();
    }
    else
    {
        // XXX Should show a save dialog or some such?

        // We can't understand the profile so bail
        judo::Element* err = SI::buildInvalidProfileNode();
        err->putAttrib("to", elem.getAttrib("from"));
        err->putAttrib("id", elem.getAttrib("id"));
        err->putAttrib("type", "error");
        *_session << err->toString().c_str();
        delete err;
        delete si;
        return;
    }
}

void GabberApp::on_discodb_filter(const DiscoDB::Item& item)
{
    const DiscoDB::Item::FeatureList& features = item.getFeatureList();
    if (std::find(features.begin(), features.end(), 
                  "jabber:iq:gateway") != features.end())
    {
        _agents.push_back(item.getJID());
    }
}

void GabberApp::on_queue_flushing()
{
    PacketQueue::queue_iterator it = _pqueue->begin();
    while (it != _pqueue->end())
    {
        PacketQueue::queue_iterator next = it;
        ++next;

        if (it->type == "PresenceRequest")
        {
            _pqueue->pop(it);
        }

        it = next;
    }
}

GabberApp& GabberApp::getSingleton(void)
{
    return Singleton<GabberApp>::getSingleton();
}

